@@ -1,17 +1,20 @@
 from .dabpumps_api import (
     DabPumpsApi, 
+    DabPumpsApiAuthError, 
+    DabPumpsApiDataError, 
+    DabPumpsApiError, 
+)
+from .dabpumps_data import (
+    DabPumpsUserRole,
+    DabPumpsParamType,
     DabPumpsInstall,
     DabPumpsDevice,
     DabPumpsConfig,
     DabPumpsParams,
     DabPumpsStatus,
-    DabPumpsApiAuthError, 
-    DabPumpsApiRightsError,
-    DabPumpsApiDataError, 
-    DabPumpsApiError, 
-    DabPumpsApiHistoryItem, 
-    DabPumpsApiHistoryDetail,
-    DabPumpsRet,
+    DabPumpsHistoryItem,
+    DabPumpsHistoryDetail,
+    DabPumpsDictFactory,
 )
 
 # for unit tests
@@ -19,3 +22,6 @@ from  .dabpumps_client import (
     DabPumpsClient_Httpx, 
     DabPumpsClient_Aiohttp,
 )
+from .dabpumps_api import (
+    DabPumpsLogin,
+)
