@@ -1,21 +1,24 @@
 import asyncio
 import copy
+from datetime import datetime
 import logging
 import pytest
 import pytest_asyncio
 
 from aiodabpumps import (
     DabPumpsApi,
+    DabPumpsApiAuthError,
+    DabPumpsApiError, 
     DabPumpsInstall,
     DabPumpsDevice,
     DabPumpsConfig,
     DabPumpsParams,
     DabPumpsStatus,
-    DabPumpsApiAuthError,
-    DabPumpsApiRightsError, 
-    DabPumpsApiError, 
-    DabPumpsApiHistoryItem, 
-    DabPumpsApiHistoryDetail,
+    DabPumpsParamType,
+    DabPumpsUserRole,
+    DabPumpsHistoryItem, 
+    DabPumpsHistoryDetail,
+    DabPumpsLogin,
 )
 
 from . import TEST_USERNAME, TEST_PASSWORD
@@ -48,16 +51,27 @@ async def context():
     # cleanup
     await ctx.cleanup()
 
+
 @pytest.mark.asyncio
 @pytest.mark.usefixtures("context")
 @pytest.mark.parametrize(
-    "name, usr, pwd, exp_except",
+    "name, method, usr, pwd, exp_except",
     [
-        ("login ok",   TEST_USERNAME, TEST_PASSWORD, None),
-        ("login fail", "dummy_usr",   "wrong_pwd",   DabPumpsApiAuthError),
+        ("ok",   'Any',           TEST_USERNAME, TEST_PASSWORD, None),
+        ("ok",   'H2D_app',       TEST_USERNAME, TEST_PASSWORD, None),
+        ("ok",   'DabLive_app_0', TEST_USERNAME, TEST_PASSWORD, None),
+        ("ok",   'DabLive_app_1', TEST_USERNAME, TEST_PASSWORD, None),
+        ("ok",   'DConnect_app',  TEST_USERNAME, TEST_PASSWORD, None),
+        ("ok",   'DConnect_web',  TEST_USERNAME, TEST_PASSWORD, None),
+        ("fail", 'Any',           "dummy_usr",   "wrong_pwd",   DabPumpsApiAuthError),
+        ("fail", 'H2D_app',       "dummy_usr",   "wrong_pwd",   DabPumpsApiAuthError),
+        ("fail", 'DabLive_app_0', "dummy_usr",   "wrong_pwd",   DabPumpsApiAuthError),
+        ("fail", 'DabLive_app_1', "dummy_usr",   "wrong_pwd",   DabPumpsApiAuthError),
+        ("fail", 'DConnect_app',  "dummy_usr",   "wrong_pwd",   DabPumpsApiAuthError),
+        ("fail", 'DConnect_web',  "dummy_usr",   "wrong_pwd",   DabPumpsApiAuthError),
     ]
 )
-async def test_login(name, usr, pwd, exp_except, request):
+async def test_login(name, method, usr, pwd, exp_except, request):
     context = request.getfixturevalue("context")
     assert context.api is None
 
@@ -67,9 +81,32 @@ async def test_login(name, usr, pwd, exp_except, request):
     if exp_except is None:
         assert context.api.login_method is None
 
-        await context.api.async_login()
+        match method:
+            case 'Any':
+                await context.api.async_login()
+
+                assert context.api.login_method is not None
+                
+            case 'H2D_app':
+                await context.api._async_login_h2d_app()
+
+                assert context.api._access_token is not None
+                assert context.api._access_expiry > datetime.min
+                assert context.api._refresh_token is not None
+                assert context.api._refresh_expiry > datetime.min
+
+            case 'DabLive_app_0':
+                await context.api._async_login_dablive_app(isDabLive=0)
+
+            case 'DabLive_app_1':
+                await context.api._async_login_dablive_app(isDabLive=1)
+
+            case 'DConnect_app':
+                await context.api._async_login_dconnect_app()
+
+            case 'DConnect_web':
+                await context.api._async_login_dconnect_web()
 
-        assert context.api.login_method is not None
         assert context.api.install_map is not None
         assert context.api.device_map is not None
         assert context.api.config_map is not None
@@ -83,7 +120,13 @@ async def test_login(name, usr, pwd, exp_except, request):
 
     else:
         with pytest.raises(exp_except):
-            await context.api.async_login()
+            match method:
+                case 'Any':             await context.api.async_login()
+                case 'H2D_app':         await context.api._async_login_h2d_app()
+                case 'DabLive_app_0':   await context.api._async_login_dablive_app(isDabLive=0)
+                case 'DabLive_app_1':   await context.api._async_login_dablive_app(isDabLive=1)
+                case 'DConnect_app':    await context.api._async_login_dconnect_app()
+                case 'DConnect_web':    await context.api._async_login_dconnect_web()
 
 
 @pytest.mark.asyncio
@@ -99,7 +142,7 @@ async def test_login_seq(name, usr, pwd, exp_except, request):
     assert context.api is None
 
     # First call with wrong pwd
-    context.api = DabPumpsApi(usr, pwd+"xxx")
+    context.api = DabPumpsApi(usr, "wrong_pwd")
     assert context.api.closed == False
     assert context.api.login_method is None
 
@@ -134,19 +177,39 @@ async def test_login_seq(name, usr, pwd, exp_except, request):
 @pytest.mark.asyncio
 @pytest.mark.usefixtures("context")
 @pytest.mark.parametrize(
-    "name, loop, exp_except",
+    "name, method, loop, exp_except",
     [
-        ("data ok", 0, None),
-        # ("data loop", 200, None),
+        ("ok",  'Auto',                      0, None),
+        ("ok",  DabPumpsLogin.H2D_APP,       0, None),
+        ("ok",  DabPumpsLogin.DABLIVE_APP_0, 0, None),
+        ("ok",  DabPumpsLogin.DABLIVE_APP_1, 0, None),
+        ("ok",  DabPumpsLogin.DCONNECT_APP,  0, None),
+        ("ok",  DabPumpsLogin.DCONNECT_WEB,  0, None),
+        ("24h", "Auto",                      24*60, None),    # Run 1 full day
+        ("24h", DabPumpsLogin.H2D_APP,       24*60, None),    # Run 1 full day
+        ("24h", DabPumpsLogin.DABLIVE_APP_1, 24*60, None),    # Run 1 full day
+        ("24h", DabPumpsLogin.DCONNECT_APP,  24*60, None),    # Run 1 full day
+        ("24h", DabPumpsLogin.DCONNECT_WEB,  24*60, None),    # Run 1 full day
     ]
 )
-async def test_get_data(name, loop, exp_except, request):
+async def test_get_data(name, method, loop, exp_except, request):
     context = request.getfixturevalue("context")
     context.api = DabPumpsApi(TEST_USERNAME, TEST_PASSWORD)
     assert context.api.closed == False
 
+    # Try set diagnostics callback function
+    context.api.set_diagnostics(lambda context,item,detail,data: None)
+
     # Login
-    await context.api.async_login()
+    match method:
+        case 'Auto':                        await context.api.async_login()
+        case DabPumpsLogin.H2D_APP:         await context.api._async_login_h2d_app()
+        case DabPumpsLogin.DABLIVE_APP_0:   await context.api._async_login_dablive_app(isDabLive=0)
+        case DabPumpsLogin.DABLIVE_APP_1:   await context.api._async_login_dablive_app(isDabLive=1)
+        case DabPumpsLogin.DCONNECT_APP:    await context.api._async_login_dconnect_app()
+        case DabPumpsLogin.DCONNECT_WEB:    await context.api._async_login_dconnect_web()
+
+    login_method_org = context.api.login_method
 
     # Get install list
     await context.api.async_fetch_install_list()
@@ -161,8 +224,8 @@ async def test_get_data(name, loop, exp_except, request):
         assert install.id is not None    
         assert install.name is not None  
 
-    # Get install details (just for the first install)
-    await context.api.async_fetch_install(install_id)
+    # Get install details, config metadata and initial statuses (just for the first install)
+    await context.api.async_fetch_install_details(install_id)
 
     assert context.api.device_map is not None
     assert type(context.api.device_map) is dict
@@ -197,13 +260,28 @@ async def test_get_data(name, loop, exp_except, request):
             assert type(param) is DabPumpsParams
             assert param.key is not None
 
-    for idx in range(loop+1):
-        # Get device statusses
-        try:
-            # Check cookie and re-login if needed
-            await context.api.async_login()
+            assert context.api.status_map is not None
+            assert type(context.api.status_map) is dict
+            assert len(context.api.status_map) > 0
 
-            await context.api.async_fetch_device_statusses(device_serial)
+    for status_id,status in context.api.status_map.items():
+        assert type(status_id) is str
+        assert type(status) is DabPumpsStatus
+        assert status.serial is not None
+        assert status.key is not None
+        assert status.name is not None
+
+    counter_success: int = 0
+    counter_fail: int = 0
+    reason_fail: dict[str,int] = {}
+    for idx in range(1,loop+1):
+        # Get fresh device statuses
+        try:
+            # Check access-token and refresh or re-login if needed
+            await context.api.async_login()
+            assert login_method_org == context.api.login_method
+
+            await context.api.async_fetch_install_statuses(install_id)
 
             assert context.api.status_map is not None
             assert type(context.api.status_map) is dict
@@ -215,38 +293,75 @@ async def test_get_data(name, loop, exp_except, request):
                 assert status.serial is not None
                 assert status.key is not None
                 assert status.name is not None
+
+            counter_success += 1
         
-        except DabPumpsApiRightsError:
-            await context.api.async_logout()
-        except:
-            continue
+        except Exception as ex:
+            counter_fail += 1
+            reason = str(ex)
+            reason_fail[reason] = reason_fail[reason]+1 if reason in reason_fail else 1
+            _LOGGER.warning(f"Fail: {ex}")
 
         if loop:
+            # Simulate failure to recover from
+            #if idx % 6 == 0:
+            #    await context.api._async_logout("simulate failure")
+            #elif idx % 3 == 0:
+            #    await context.api._async_logout("login force refresh", DabPumpsLogin.ACCESS_TOKEN)
+
+            if method != "Auto":
+                context.api._login_method = method
+
+            _LOGGER.debug(f"Loop test, {idx} of {loop} (success={counter_success}, fail={counter_fail})")
             await asyncio.sleep(60)
-            _LOGGER.debug(f"Loop test, {idx} of {loop}")
+
+    _LOGGER.info(f"Fail summary after {loop} loops:")
+    for reason,count in reason_fail.items():
+        _LOGGER.info(f"  {count}x {reason}")
+
+    assert counter_fail == 0
 
 
 @pytest.mark.asyncio
 @pytest.mark.usefixtures("context")
 @pytest.mark.parametrize(
-    "name, key, code, exp_code, exp_except",
+    "method, key, codes, exp_code, exp_except",
     [
-        ("set PowerShowerBoost 30", "PowerShowerBoost", "30", "30", None),
-        ("set PowerShowerBoost 20", "PowerShowerBoost", "20", "20", None),
-        ("set PowerShowerDuration 360", "PowerShowerDuration", "360", "360", None),
-        ("set PowerShowerDuration 300", "PowerShowerDuration", "300", "300", None),
-        ("set SleepModeEnable on", "SleepModeEnable", "1", "1", None),
-        ("set SleepModeEnable off", "SleepModeEnable", "0", "0", None),
-        ("set RF_EraseHistoricalFault", "RF_EraseHistoricalFault", "1", "0", None), # Falls back to 0 after STATUS_UPDATE_HOLD
+        ('Auto',                      "PowerShowerBoost",        ["20","30"],   "=", None),
+        ('Auto',                      "PowerShowerDuration",     ["300","360"], "=", None),
+        ('Auto',                      "SleepModeEnable",         ["0", "1"],    "=", None),
+        ('Auto',                      "RF_EraseHistoricalFault", ["1"],         "0", None), # Falls back to 0 after STATUS_UPDATE_HOLD
+        (DabPumpsLogin.H2D_APP,       "PowerShowerBoost",        ["20","30"],   "=", None),
+        (DabPumpsLogin.H2D_APP,       "PowerShowerDuration",     ["300","360"], "=", None),
+        (DabPumpsLogin.H2D_APP,       "SleepModeEnable",         ["0", "1"],    "=", None),
+        (DabPumpsLogin.H2D_APP,       "RF_EraseHistoricalFault", ["1"],         "0", None), # Falls back to 0 after STATUS_UPDATE_HOLD
+        (DabPumpsLogin.DABLIVE_APP_1, "PowerShowerBoost",        ["20","30"],   "=", None),
+        (DabPumpsLogin.DABLIVE_APP_1, "PowerShowerDuration",     ["300","360"], "=", None),
+        (DabPumpsLogin.DABLIVE_APP_1, "SleepModeEnable",         ["0", "1"],    "=", None),
+        (DabPumpsLogin.DABLIVE_APP_1, "RF_EraseHistoricalFault", ["1"],         "0", None), # Falls back to 0 after STATUS_UPDATE_HOLD
+        (DabPumpsLogin.DCONNECT_APP,  "PowerShowerBoost",        ["20","30"],   "=", None),
+        (DabPumpsLogin.DCONNECT_APP,  "PowerShowerDuration",     ["300","360"], "=", None),
+        (DabPumpsLogin.DCONNECT_APP,  "SleepModeEnable",         ["0", "1"],    "=", None),
+        (DabPumpsLogin.DCONNECT_APP,  "RF_EraseHistoricalFault", ["1"],         "0", None), # Falls back to 0 after STATUS_UPDATE_HOLD
+        (DabPumpsLogin.DCONNECT_WEB,  "PowerShowerBoost",        ["20","30"],   "=", None),
+        (DabPumpsLogin.DCONNECT_WEB,  "PowerShowerDuration",     ["300","360"], "=", None),
+        (DabPumpsLogin.DCONNECT_WEB,  "SleepModeEnable",         ["0", "1"],    "=", None),
+        (DabPumpsLogin.DCONNECT_WEB,  "RF_EraseHistoricalFault", ["1"],         "0", None), # Falls back to 0 after STATUS_UPDATE_HOLD
     ]
 )
-async def test_set_data(name, key, code, exp_code, exp_except, request):
+async def test_set_data(method, key, codes, exp_code, exp_except, request):
     context = request.getfixturevalue("context")
     context.api = DabPumpsApi(TEST_USERNAME, TEST_PASSWORD)
     assert context.api.closed == False
 
     # Login
-    await context.api.async_login()
+    match method:
+        case 'Auto':                        await context.api.async_login()
+        case DabPumpsLogin.H2D_APP:         await context.api._async_login_h2d_app()
+        case DabPumpsLogin.DABLIVE_APP_0:   await context.api._async_login_dablive_app(isDabLive=0)
+        case DabPumpsLogin.DABLIVE_APP_1:   await context.api._async_login_dablive_app(isDabLive=1)
+        case DabPumpsLogin.DCONNECT_APP:    await context.api._async_login_dconnect_app()
+        case DabPumpsLogin.DCONNECT_WEB:    await context.api._async_login_dconnect_web()
 
     # Get install list
     await context.api.async_fetch_install_list()
@@ -255,38 +370,53 @@ async def test_set_data(name, key, code, exp_code, exp_except, request):
     assert type(context.api.install_map) is dict
     assert len(context.api.install_map) > 0
 
-    # Get install details and metadata
+    # Get install details, metadata and initial statuses
     for install_id in context.api.install_map:
-        await context.api.async_fetch_install(install_id)
-
-    # Get device statusses
-    for device_serial in context.api.device_map:
-        await context.api.async_fetch_device_statusses(device_serial)
+        await context.api.async_fetch_install_details(install_id)
 
+    # Find current code and value and find a new code to change into
     status = next( (status for status in context.api.status_map.values() if status.key==key), None)
     assert status is not None
 
-    changed = await context.api.async_change_device_status(status.serial, status.key, code=code)
+    old_code = status.code
+    new_code = next( (code for code in codes if code != old_code), None )
 
-    # Do immediate test of changed value. 
+    # Change device status and do immediate test of changed value. 
     # We hold the changed value while the backend is processing the change.
+    changed = await context.api.async_change_device_status(status.serial, status.key, code=new_code)
     if changed:
-        await context.api.async_fetch_device_statusses(status.serial)
+        await context.api.async_fetch_install_statuses(install_id)
 
         status = next( (status for status in context.api.status_map.values() if status.key==key), None)
-        assert status.code == code
+        assert status.code == new_code
         assert status.update_ts is not None
+        _LOGGER.debug(f"Found value changed from {old_code} to {new_code}")
 
         # Wait until the backend has processed the change and test again
+        _LOGGER.debug(f"Wait for DAB Servers to process the change")
         await asyncio.sleep(40)
+        await context.api.async_login()
 
-    # Test (either not changed or after change has been processed by backend)
-    await context.api.async_fetch_device_statusses(status.serial)
+    # Test after change has been processed by backend
+    await context.api.async_fetch_install_statuses(install_id)
 
     status = next( (status for status in context.api.status_map.values() if status.key==key), None)
-    assert status.code == exp_code
+    assert status.code == new_code if exp_code == "=" else exp_code
     assert status.update_ts is None
 
+    _LOGGER.debug(f"Found value still changed from {old_code} to {new_code}")
+
+    # Change back to original value and do immediate test of changed value
+    changed = await context.api.async_change_device_status(status.serial, status.key, code=old_code)
+    if changed:
+        await context.api.async_fetch_install_statuses(install_id)
+
+        status = next( (status for status in context.api.status_map.values() if status.key==key), None)
+        assert status.code == old_code
+        assert status.update_ts is not None
+
+        _LOGGER.debug(f"Found value changed back from {new_code} to {old_code}")
+
 
 @pytest.mark.asyncio
 @pytest.mark.usefixtures("context")
@@ -355,10 +485,10 @@ async def config_map():
             label = 'test label',
             description = 'test description',
             meta_params = {
-                "KEY_ENUM":  DabPumpsParams(key='KEY_ENUM',  type='enum',    unit=None, weight=None, values={'1':'one', '2':'two', '3':'three'}, min=1, max=3, family='f', group='g', view='CSIR', change='', log='', report=''),
-                "KEY_FLOAT": DabPumpsParams(key='KEY_FLOAT', type='measure', unit='F',  weight=0.1,  values=None, min=0, max=1,  family='f', group='g', view='CSIR', change='', log='', report=''),
-                "KEY_INT":   DabPumpsParams(key='KEY_INT',   type='measure', unit='I',  weight=1,    values=None, min=0, max=10, family='f', group='g', view='CSIR', change='', log='', report=''),
-                "KEY_LABEL": DabPumpsParams(key='KEY_LABEL', type='label',   unit='',   weight=None, values=None, min=0, max=0,  family='f', group='g', view='CSIR', change='', log='', report=''),
+                "KEY_ENUM":  DabPumpsParams(key='KEY_ENUM',  type=DabPumpsParamType.ENUM,    unit=None, weight=None, values={'1':'one', '2':'two', '3':'three'}, min=1, max=3, family='f', group='g', view='CSIR', change='', log='', report=''),
+                "KEY_FLOAT": DabPumpsParams(key='KEY_FLOAT', type=DabPumpsParamType.MEASURE, unit='F',  weight=0.1,  values=None, min=0, max=1,  family='f', group='g', view='CSIR', change='', log='', report=''),
+                "KEY_INT":   DabPumpsParams(key='KEY_INT',   type=DabPumpsParamType.MEASURE, unit='I',  weight=1,    values=None, min=0, max=10, family='f', group='g', view='CSIR', change='', log='', report=''),
+                "KEY_LABEL": DabPumpsParams(key='KEY_LABEL', type=DabPumpsParamType.LABEL,   unit='',   weight=None, values=None, min=0, max=0,  family='f', group='g', view='CSIR', change='', log='', report=''),
             }
         ),
     }
@@ -494,11 +624,11 @@ async def test_status(name, serial, key, exp_code, exp_value, exp_unit, request)
         ("device unknown", 'SERIAL_XX', 'KEY_ENUM', False, None, None, None),
         ("device unknown", 'SERIAL_XX', 'KEY_ENUM', False, None, None, None),
         ("key unknown", 'SERIAL', 'KEY_XX', False, None, None, None),
-        ("enum ok", "SERIAL", 'KEY_ENUM', False, 'enum', {'1':'one', '2':'two', '3':'three'}, None),
-        ("enum ok", "SERIAL", 'KEY_ENUM', True, 'enum', {'1':'een', '2':'twee', '3':'drie'}, None),
-        ("float ok", "SERIAL", 'KEY_FLOAT', False, 'measure', None, 'F'),
-        ("int ok", "SERIAL", 'KEY_INT', False, 'measure', None, 'I'),
-        ("label ok", "SERIAL", 'KEY_LABEL', False, 'label', None, ''),
+        ("enum ok", "SERIAL", 'KEY_ENUM', False, DabPumpsParamType.ENUM, {'1':'one', '2':'two', '3':'three'}, None),
+        ("enum ok", "SERIAL", 'KEY_ENUM', True, DabPumpsParamType.ENUM, {'1':'een', '2':'twee', '3':'drie'}, None),
+        ("float ok", "SERIAL", 'KEY_FLOAT', False, DabPumpsParamType.MEASURE, None, 'F'),
+        ("int ok", "SERIAL", 'KEY_INT', False, DabPumpsParamType.MEASURE, None, 'I'),
+        ("label ok", "SERIAL", 'KEY_LABEL', False, DabPumpsParamType.LABEL, None, ''),
     ]
 )
 async def test_metadata(name, serial, key, translate, exp_type, exp_values, exp_unit, request):
